from datetime import date, datetime
from typing import List
from typing import Optional

from pydantic import BaseModel
from pydantic import Field
from pydantic import root_validator
from pydantic import validator

from ..common import Pincode
from ..validators import normalize_str
from .enums import EmploymentStatus
from .enums import EnquiryReason
from .enums import FinancePurpose
from .enums import Gender
from .enums import MaritalStatus
from .enums import PhoneNumberType
from .enums import State
from .enums import ConsentType


class PhoneNumber(BaseModel):
    phone_number: str = Field(max_length=20)
    extension: Optional[str] = Field(max_length=10)
    number_type: Optional[PhoneNumberType]

    _normalize_num = validator("phone_number", allow_reuse=True, pre=True)(
        normalize_str
    )


class Identity(BaseModel):
    id_number: str
    issue_date: Optional[date]
    expiration_date: Optional[date]

    _normalize_str = validator("id_number", allow_reuse=True, pre=True)(normalize_str)

    @root_validator
    def check_issue__before_expiration(cls, values: dict):
        if values.get("issue_date") and values.get("expiration_date"):
            assert (
                values["issue_date"] < values["expiration_date"]
            ), "issue date must be before expiration date"

        return values


class Address(BaseModel):
    address_line_1: Optional[str] = Field(max_length=200)
    address_line_2: Optional[str] = Field(max_length=200)
    address_line_3: Optional[str] = Field(max_length=200)
    city: Optional[str] = Field(max_length=50)
    landmark: Optional[str] = Field(max_length=40)
    state: Optional[State]
    pincode: Pincode

    _normalize_strs = validator("address_line_1", "city", allow_reuse=True, pre=True)(
        normalize_str
    )


class Consent(BaseModel):
    consent_flag: bool = Field()
    consent_timestamp: Optional[datetime] = Field()
    consent_ip: str = Field()
    consentMessageId: ConsentType


class Applicant(BaseModel):
    first_name: str = Field(max_length=40)
    middle_name_1: Optional[str] = Field(max_length=40)
    middle_name_2: Optional[str] = Field(max_length=40)
    middle_name_3: Optional[str] = Field(max_length=40)
    last_name: str = Field(max_length=60)
    gender: Optional[Gender]
    date_of_birth: Optional[date]
    email: Optional[str] = Field(max_length=80)
    phone_number: Optional[PhoneNumber]
    mobile_number: Optional[str] = Field(max_length=20)
    income: Optional[str] = Field(max_length=15)
    marital_status: Optional[MaritalStatus]
    employment_status: Optional[EmploymentStatus]
    time_with_employer: Optional[str] = Field(max_length=3)
    number_of_major_credit_cards_held: Optional[int] = Field(lt=1000)
    pan: Optional[Identity]
    passport: Optional[Identity]
    voter_id: Optional[Identity]
    drivers_licence: Optional[Identity]
    ration_card: Optional[Identity]
    uid: Optional[Identity]
    addresses: Optional[List[Address]] = Field(min_items=1, max_items=2)
    alternate_flow: Optional[str] = Field()
    consent: Optional[Consent]

    _normalize_strs = validator("first_name", "last_name", allow_reuse=True, pre=True)(
        normalize_str
    )


class ExperianRequest(BaseModel):
    ft_reference_number: Optional[str] = Field(max_length=6)
    customer_reference_number: Optional[str] = Field(max_length=12)
    enquiry_reason: Optional[EnquiryReason]
    finance_purpose: Optional[FinancePurpose]
    amount_financed: Optional[int]
    duration_of_agreement: Optional[int] = Field(lt=1000)
    applicant: Applicant



from datetime import datetime
from typing import Optional
from pydantic import BaseModel, Field


from mercuryclient.types.common import NormalizedString
from mercuryclient.types.payments.enums import (
    PlanType,
    SubscriptionActionType,
    EmandateMethod,
    AuthType,
    Currency,
    TokenRequestType,
)


class CustomerDetails(BaseModel):

    """
    Customer Details
    """

    name: NormalizedString = Field(max_length=100)
    email: NormalizedString = Field(max_length=100)
    contact: NormalizedString = Field(max_length=15)


class NotesField(BaseModel):
    """
    Notes Field Details
    """

    purpose: Optional[NormalizedString] = Field(max_length=250)
    note_key_1: Optional[NormalizedString] = Field(max_length=250)
    note_key_2: Optional[NormalizedString] = Field(max_length=250)


class BankDetails(BaseModel):
    """
    Bank Details
    """

    account_holder_name: Optional[NormalizedString] = Field()
    account_number: Optional[NormalizedString] = Field()
    account_type: Optional[NormalizedString] = Field()
    account_bank_code: Optional[NormalizedString] = Field()
    beneficiary_name: Optional[NormalizedString] = Field()
    ifsc_code: Optional[NormalizedString] = Field()


class ItemField(BaseModel):
    """
    Item Field Details
    """

    name: NormalizedString = Field(max_length=80)
    amount: float = Field()
    currency: NormalizedString = Field(max_length=8)
    description: NormalizedString = Field(max_length=250)


class QRCodeDetails(BaseModel):
    """
    QR Code details
    """

    type: NormalizedString = Field(max_length=20)
    name: NormalizedString = Field(max_length=80)
    usage: NormalizedString = Field(max_length=25)
    fixed_amount: bool = Field()
    payment_amount: float = Field()
    description: NormalizedString = Field(max_length=250)
    notes: Optional[NotesField]


class QRCodeClosure(BaseModel):
    """
    QR Code Closure details
    """

    qr_id: NormalizedString = Field(max_length=100)


class PlanDetails(BaseModel):
    plan_name: Optional[NormalizedString] = Field(max_length=20)
    plan_type: Optional[PlanType]
    plan_currency: Optional[NormalizedString] = Field(max_length=20, default="INR")
    max_amount: Optional[float]
    plan_note: Optional[NormalizedString] = Field(max_length=20)
    plan_recurring_amount: Optional[float] = Field()


class CreatePlan(PlanDetails):
    """
    Payment Plan Details
    """

    period: NormalizedString = Field(max_length=20)
    interval: int = Field()
    item: Optional[ItemField]

    plan_id: Optional[NormalizedString] = Field(max_length=20)


class SubscriptionCreation(BaseModel):

    """
    Subscription Creation Details
    """

    plan_id: Optional[NormalizedString] = Field(max_length=100)
    subscription_id: Optional[NormalizedString] = Field(max_length=100)

    total_count: Optional[int] = Field()
    quantity: Optional[int] = Field()
    customer_notify: Optional[bool] = Field(default=False)
    customer_details: Optional[CustomerDetails]
    expiry_time: Optional[datetime] = Field()
    first_charge_time: Optional[datetime] = Field()
    note: Optional[NormalizedString] = Field(max_length=100)
    bank_details: Optional[BankDetails]
    plan_details: Optional[PlanDetails]


class SubscriptionFetch(BaseModel):

    """
    Serializer for subscription fetch
    """

    subscription_id: NormalizedString = Field(max_length=100)


class Notify(BaseModel):

    """
    Notify Details
    """

    sms: bool = Field()
    email: bool = Field()


class PaymentGateway(BaseModel):
    """
    Payment Gateway Link Details
    """

    amount: float = Field()
    currency: NormalizedString = Field(max_length=10)
    accept_partial: bool = Field()
    description: NormalizedString = Field(max_length=250)
    notes: Optional[NotesField]
    customer: CustomerDetails
    notify: Optional[Notify]
    reminder_enable: bool = Field()


class PaymentLinkClosure(BaseModel):
    """
    Payment Link Closure Details
    """

    link_id: NormalizedString = Field(max_length=100)


class SubscriptionCharge(BaseModel):

    """
    Serializer for subscription Charge
    """

    subscription_id: NormalizedString = Field(max_length=100)
    payment_id: NormalizedString = Field(max_length=100)
    payment_amount: float
    payment_schedule_date: datetime


class SubscriptionManage(BaseModel):

    """
    Serializer for subscription manage
    """

    subscription_id: NormalizedString = Field(max_length=40)
    action: SubscriptionActionType
    next_scheduled_time: Optional[NormalizedString]
    plan_id: Optional[NormalizedString]
    cancel_at_cycle_end: Optional[int] = Field()
    pause_at: Optional[NormalizedString] = Field(max_length=250)
    resume_at: Optional[NormalizedString] = Field(max_length=250)


class EmandateSubscription(BaseModel):
    method: EmandateMethod
    auth_type: AuthType
    max_amount: int = Field()
    expire_at: Optional[int] = Field()
    bank_account: BankDetails


class EmandateRegistration(BaseModel):
    customer: CustomerDetails
    type: NormalizedString = Field()
    amount: int = Field()
    currency: Currency
    description: NormalizedString = Field()
    subscription_registration: EmandateSubscription
    sms_notify: Optional[bool] = Field()
    email_notify: Optional[bool] = Field()
    expire_by: Optional[int] = Field()
    receipt: Optional[NormalizedString] = Field()
    notes: Optional[NotesField]


class EmandatePaymentToken(BaseModel):
    payment_cust_id: NormalizedString = Field()
    request_type: TokenRequestType


class EmandateOrder(BaseModel):
    amount: int = Field()
    currency: Currency
    payment_capture: bool = Field()
    receipt: Optional[NormalizedString] = Field()
    notes: Optional[NotesField]


class EmandateRecurringPayment(BaseModel):
    email: NormalizedString = Field()
    contact: int = Field()
    currency: Currency
    amount: int = Field()
    order_id: NormalizedString = Field()
    customer_id: NormalizedString = Field()
    token: NormalizedString = Field()
    recurring: bool = Field()
    description: Optional[NormalizedString] = Field()
    notes: Optional[NotesField]



from enum import Enum


class Endpoints:
    api_endpoints = {
        "generate_qr_code": "generate/qr/code",
        "plan_creation": "plan/create",
        "subscription_creation": "subscription/create",
        "subscription_fetch": "subscription/fetch",
        "payment_link_creation": "payment/link/create",
        "cancel_payment_link": "payment/link/cancel",
        "close_qr_code": "qr/code/closure",
        "charge_subscription": "subscription/charge",
        "manage_subscription": "subscription/manage",
        "emandate_registration": "emandate/registration/create",
        "emandate_payment_token": "emandate/payment/token",
        "emandate_order_creation": "emandate/order/creation",
        "emandate_recurring_payments": "emandate/recurring/payments",
    }


class PlanType(Enum):
    PERIODIC = "PERIODIC"
    ON_DEMAND = "ON_DEMAND"


class SubscriptionActionType(Enum):
    CANCEL = "CANCEL"
    PAUSE = "PAUSE"
    ACTIVATE = "ACTIVATE"
    CHNAGE_PLAN = "CHANGE_PLAN"
    RESUME = "RESUME"


class EmandateMethod(Enum):
    EMANDATE = "emandate"


class AuthType(Enum):
    NETBANKING = "netbanking"
    AADHAAR = "aadhaar"
    DEBITCARD = "debitcard"


class Currency(Enum):
    INR = "INR"


class TokenRequestType(Enum):
    PAYMENTID = "PAYMENT_ID"
    CUSTOMERID = "CUSTOMER_ID"



import time
from typing import Tuple
import json

from mercuryclient.types.payments.enums import Endpoints
from mercuryclient.types.payments.request import (
    QRCodeDetails,
    CreatePlan,
    SubscriptionCreation,
    SubscriptionFetch,
    PaymentGateway,
    SubscriptionCharge,
    SubscriptionManage,
    PaymentLinkClosure,
    QRCodeClosure,
    EmandateRegistration,
    EmandatePaymentToken,
    EmandateOrder,
    EmandateRecurringPayment,
)


class PaymentsMixin:
    """
    Mixin for initiating payment
    """

    def initiate_payment_request(
        self, profile, provider, request_data, request_type
    ) -> str:
        api = "api/v1/payments/"
        try:
            endpoint = Endpoints.api_endpoints[request_type]
        except ValueError:
            raise Exception(f"{endpoint} is not a valid request_type")
        url = f"{api}{endpoint}"

        request_dict = json.loads(request_data.json(exclude_unset=True))
        request_dict["profile"] = profile
        request_dict["provider"] = provider
        request_id, r = self._post_json_http_request(
            url, data=request_dict, send_request_id=True, add_bearer_token=True
        )
        if r.status_code == 201:
            return request_id

        try:
            response_json = r.json()
        except Exception:
            response_json = {}
        raise Exception(
            "Error while sending payment request. Status: {}, Response is {}".format(
                r.status_code, response_json
            )
        )

    def get_payment_response(self, request_id: str) -> Tuple[str, dict]:
        api = "api/v1/payments/result"

        request_id, r = self._get_json_http_request(
            api,
            headers={"X-Mercury-Request-Id": request_id},
            send_request_id=False,
            add_bearer_token=True,
        )

        if r.status_code == 200:
            result = r.json()
            return request_id, result

        try:
            response_json = r.json()
        except Exception:
            response_json = {}
        raise Exception(
            "Error getting Payment Operation result. Status: {}, Response is {}".format(
                r.status_code, response_json
            )
        )

    def get_payment_result(
        self, mercury_request_id: str, max_attempts: int = 5, retry_backoff: int = 5
    ):
        """Retry mechanism for okyc response

        raises:
            Exception: After max attempts, response till remains in progress

        returns mercury_request_id,response from the vendor

        """

        for attempts in range(max_attempts):
            time.sleep(retry_backoff)
            mercury_request_id, result = self.get_payment_response(mercury_request_id)
            if result.get("status") != "IN_PROGRESS":
                return mercury_request_id, result

            retry_backoff *= 2

        raise Exception("Error getting Payment Operation result. Status: IN_PROGRESS")

    def generate_qr_code(
        self,
        request_obj: QRCodeDetails,
        profile: str,
        provider: str,
        max_attempts: int = 5,
        retry_backoff: int = 5,
    ):
        """Initiates the request for creating qr code.

        returns: mercury_request_id,qr code response from vendor
        """

        mercury_request_id = self.initiate_payment_request(
            profile=profile,
            provider=provider,
            request_data=request_obj,
            request_type="generate_qr_code",
        )
        return self.get_payment_result(mercury_request_id, max_attempts, retry_backoff)

    def qr_code_closure(
        self,
        request_obj: QRCodeClosure,
        profile: str,
        provider: str,
        max_attempts: int = 5,
        retry_backoff: int = 5,
    ):
        """Initiates the request for closing a qr code.

        returns: mercury_request_id,qr code response from vendor
        """

        mercury_request_id = self.initiate_payment_request(
            profile=profile,
            provider=provider,
            request_data=request_obj,
            request_type="close_qr_code",
        )
        return self.get_payment_result(mercury_request_id, max_attempts, retry_backoff)

    def generate_payment_plan(
        self,
        request_obj: CreatePlan,
        profile: str,
        provider: str,
        max_attempts: int = 5,
        retry_backoff: int = 5,
    ):
        """Initiates the request for payment plan.

        returns: mercury_request_id,plan id inside the response from vendor
        """

        mercury_request_id = self.initiate_payment_request(
            profile=profile,
            provider=provider,
            request_data=request_obj,
            request_type="plan_creation",
        )
        return self.get_payment_result(mercury_request_id, max_attempts, retry_backoff)

    def create_payment_subscription(
        self,
        request_obj: SubscriptionCreation,
        profile: str,
        provider: str,
        max_attempts: int = 5,
        retry_backoff: int = 5,
    ):
        """Initiates the request creating a payment subscription.

        returns: mercury_request_id,otp generation response from vendor
        """

        mercury_request_id = self.initiate_payment_request(
            profile=profile,
            provider=provider,
            request_data=request_obj,
            request_type="subscription_creation",
        )
        return self.get_payment_result(mercury_request_id, max_attempts, retry_backoff)

    def fetch_subscription(
        self,
        request_obj: SubscriptionFetch,
        profile: str,
        provider: str,
        max_attempts: int = 5,
        retry_backoff: int = 5,
    ):
        """Initiates the request fetching details of an existing subscription.

        returns: mercury_request_id,otp generation response from vendor
        """

        mercury_request_id = self.initiate_payment_request(
            profile=profile,
            provider=provider,
            request_data=request_obj,
            request_type="subscription_fetch",
        )
        return self.get_payment_result(mercury_request_id, max_attempts, retry_backoff)

    def generate_payment_link(
        self,
        request_obj: PaymentGateway,
        profile: str,
        provider: str,
        max_attempts: int = 5,
        retry_backoff: int = 5,
    ):
        """Initiates the request for generating a payment gateway link.

        returns: mercury_request_id,otp generation response from vendor
        """

        mercury_request_id = self.initiate_payment_request(
            profile=profile,
            provider=provider,
            request_data=request_obj,
            request_type="payment_link_creation",
        )
        return self.get_payment_result(mercury_request_id, max_attempts, retry_backoff)

    def cancel_payment_link(
        self,
        request_obj: PaymentLinkClosure,
        profile: str,
        provider: str,
        max_attempts: int = 5,
        retry_backoff: int = 5,
    ):
        """Initiates the request the closure of a payment gateway link.

        returns: mercury_request_id,otp generation response from vendor
        """

        mercury_request_id = self.initiate_payment_request(
            profile=profile,
            provider=provider,
            request_data=request_obj,
            request_type="cancel_payment_link",
        )
        return self.get_payment_result(mercury_request_id, max_attempts, retry_backoff)

    def charge_subscription(
        self,
        request_obj: SubscriptionCharge,
        profile: str,
        provider: str,
        max_attempts: int = 5,
        retry_backoff: int = 5,
    ):
        """Initiates the request fetching details of an existing subscription.

        returns: mercury_request_id,otp generation response from vendor
        """

        mercury_request_id = self.initiate_payment_request(
            profile=profile,
            provider=provider,
            request_data=request_obj,
            request_type="charge_subscription",
        )
        return self.get_payment_result(mercury_request_id, max_attempts, retry_backoff)

    def manage_subscription(
        self,
        request_obj: SubscriptionManage,
        profile: str,
        provider: str,
        max_attempts: int = 5,
        retry_backoff: int = 5,
    ):
        """Initiates the request fetching details of an existing subscription.

        returns: mercury_request_id,otp generation response from vendor
        """

        mercury_request_id = self.initiate_payment_request(
            profile=profile,
            provider=provider,
            request_data=request_obj,
            request_type="manage_subscription",
        )
        return self.get_payment_result(mercury_request_id, max_attempts, retry_backoff)

    def emandate_registration(
        self,
        request_obj: EmandateRegistration,
        profile: str,
        provider: str,
        max_attempts: int = 5,
        retry_backoff: int = 5,
    ):
        """Initiates the request to create an emandate registration link.

        returns: mercury_request_id,otp generation response from vendor
        """

        mercury_request_id = self.initiate_payment_request(
            profile=profile,
            provider=provider,
            request_data=request_obj,
            request_type="emandate_registration",
        )
        return self.get_payment_result(mercury_request_id, max_attempts, retry_backoff)

    def emandate_payment_token(
        self,
        request_obj: EmandatePaymentToken,
        profile: str,
        provider: str,
        max_attempts: int = 5,
        retry_backoff: int = 5,
    ):
        """Initiates the request to create an emandate payment token.

        returns: mercury_request_id,otp generation response from vendor
        """

        mercury_request_id = self.initiate_payment_request(
            profile=profile,
            provider=provider,
            request_data=request_obj,
            request_type="emandate_payment_token",
        )
        return self.get_payment_result(mercury_request_id, max_attempts, retry_backoff)

    def emandate_order(
        self,
        request_obj: EmandateOrder,
        profile: str,
        provider: str,
        max_attempts: int = 5,
        retry_backoff: int = 5,
    ):
        """Initiates the request to create an emandate payment token.

        returns: mercury_request_id,otp generation response from vendor
        """

        mercury_request_id = self.initiate_payment_request(
            profile=profile,
            provider=provider,
            request_data=request_obj,
            request_type="emandate_order_creation",
        )
        return self.get_payment_result(mercury_request_id, max_attempts, retry_backoff)

    def emandate_recurring_payments(
        self,
        request_obj: EmandateRecurringPayment,
        profile: str,
        provider: str,
        max_attempts: int = 5,
        retry_backoff: int = 5,
    ):
        """Initiates the request to create an emandate payment token.

        returns: mercury_request_id,otp generation response from vendor
        """

        mercury_request_id = self.initiate_payment_request(
            profile=profile,
            provider=provider,
            request_data=request_obj,
            request_type="emandate_recurring_payments",
        )
        return self.get_payment_result(mercury_request_id, max_attempts, retry_backoff)



import json
from mercuryclient.api import MercuryApi
from unittest import TestCase, mock
from mercuryclient.types.payments.enums import Endpoints
from mercuryclient.types.payments.request import (
    QRCodeDetails,
    CreatePlan,
    SubscriptionCreation,
    SubscriptionFetch,
    PaymentGateway,
    QRCodeClosure,
    PaymentLinkClosure,
)


class PaymentsMixinTest(TestCase):
    def setUp(self):
        self.client = MercuryApi(
            {
                "username": "username",
                "password": "password",
                "url": "https://mercury-dev.esthenos.in",
            }
        )

        qr_code_data = {
            "type": "test_qr_code",
            "name": "test_name",
            "usage": "test_use",
            "fixed_amount": True,
            "payment_amount": 999,
            "description": "Pass Case",
            "notes": {"purpose": "testing"},
        }
        plan_creation_data = {
            "period": "monthly",
            "interval": 1,
            "item": {
                "name": "test_user",
                "amount": 9999,
                "currency": "INR",
                "description": "Pass Case",
            },
        }
        subscription_creation_data = {
            "plan_id": "RZP001",
            "total_count": 12,
            "quantity": 1,
            "customer_notify": 0,
        }
        subscription_fetch_data = {
            "provider": "razorpay",
            "profile": "sample_profile",
            "subscription_id": "RZPSC001",
        }
        payment_link_data = {
            "provider": "razorpay",
            "profile": "sample_profile",
            "amount": 123244,
            "currency": "INR",
            "accept_partial": False,
            "description": "Collection of loan id 21343",
            "notes": {
                "purpose": "Collection of loan id 21343",
            },
            "customer": {
                "name": "test customer",
                "email": "test@test.com",
                "contact": "9837632567",
            },
            "notify": {"sms": True, "email": True},
            "reminder_enable": True,
        }
        payment_link_closure_data = {
            "provider": "razorpay",
            "profile": "sample_profile",
            "link_id": "pymtslink12343",
        }
        qr_code_closure_data = {
            "provider": "razorpay",
            "profile": "sample_profile",
            "qr_id": "pymtsqr12343",
        }
        self.qr_code_request_obj = QRCodeDetails(**qr_code_data)
        self.plan_creation_obj = CreatePlan(**plan_creation_data)
        self.subscription_creation_obj = SubscriptionCreation(
            **subscription_creation_data
        )
        self.subscription_fetch_obj = SubscriptionFetch(**subscription_fetch_data)
        self.payment_link_obj = PaymentGateway(**payment_link_data)
        self.payment_link_closure_obj = PaymentLinkClosure(**payment_link_closure_data)
        self.qr_code_closure_obj = QRCodeClosure(**qr_code_closure_data)

        self.qr_code_request_dict = {
            **json.loads(self.qr_code_request_obj.json(exclude_unset=True)),
            **{"profile": "sample_profile"},
            **{"provider": "razorpay"},
        }
        self.qr_code_closure_dict = {
            **json.loads(self.qr_code_closure_obj.json(exclude_unset=True)),
            **{"profile": "sample_profile"},
            **{"provider": "razorpay"},
        }
        self.plan_creation_dict = {
            **json.loads(self.plan_creation_obj.json(exclude_unset=True)),
            **{"profile": "sample_profile"},
            **{"provider": "razorpay"},
        }
        self.subscription_creation_dict = {
            **json.loads(self.subscription_creation_obj.json(exclude_unset=True)),
            **{"profile": "sample_profile"},
            **{"provider": "razorpay"},
        }
        self.subscription_fetch_dict = {
            **json.loads(self.subscription_fetch_obj.json(exclude_unset=True)),
            **{"profile": "sample_profile"},
            **{"provider": "razorpay"},
        }
        self.payment_link_dict = {
            **json.loads(self.payment_link_obj.json(exclude_unset=True)),
            **{"profile": "sample_profile"},
            **{"provider": "razorpay"},
        }
        self.payment_link_closure_dict = {
            **json.loads(self.payment_link_closure_obj.json(exclude_unset=True)),
            **{"profile": "sample_profile"},
            **{"provider": "razorpay"},
        }

        self.post_api_mock = mock.patch(
            "mercuryclient.api.MercuryApi._post_json_http_request"
        ).start()
        self.addCleanup(self.post_api_mock.stop)
        self.get_api_mock = mock.patch(
            "mercuryclient.api.MercuryApi._get_json_http_request"
        ).start()
        self.addCleanup(self.get_api_mock.stop)

    def test_entire_qr_code_flow(self):
        mock_post_response = mock.MagicMock()
        mock_post_response.status_code = 201
        mock_get_response = mock.MagicMock()
        mock_get_response.status_code = 200
        mock_get_response.json = mock.MagicMock()
        mock_get_response.json.return_value = {"status": "SUCCESS"}
        self.post_api_mock.return_value = ("random_string", mock_post_response)
        self.get_api_mock.return_value = ("random_string", mock_get_response)

        response = self.client.generate_qr_code(
            self.qr_code_request_obj, "sample_profile", "razorpay"
        )

        self.post_api_mock.assert_called_with(
            "api/v1/payments/generate/qr/code",
            data=self.qr_code_request_dict,
            send_request_id=True,
            add_bearer_token=True,
        )

        self.get_api_mock.assert_called_with(
            "api/v1/payments/result",
            headers={"X-Mercury-Request-Id": "random_string"},
            send_request_id=False,
            add_bearer_token=True,
        )

        self.assertEqual(response[1]["status"], "SUCCESS")

    def test_qr_code_request_exception_raised_if_status_code_error(self):
        mock_response = mock.MagicMock()
        mock_response.status_code = 401
        self.post_api_mock.return_value = ("random_string", mock_response)
        with self.assertRaises(Exception):
            self.client.generate_qr_code(
                self.qr_code_request_obj, "test", "some_provider"
            )

    def test_entire_qr_code_closure_flow(self):
        mock_post_response = mock.MagicMock()
        mock_post_response.status_code = 201
        mock_get_response = mock.MagicMock()
        mock_get_response.status_code = 200
        mock_get_response.json = mock.MagicMock()
        mock_get_response.json.return_value = {"status": "SUCCESS"}
        self.post_api_mock.return_value = ("random_string", mock_post_response)
        self.get_api_mock.return_value = ("random_string", mock_get_response)

        response = self.client.qr_code_closure(
            self.qr_code_closure_obj, "sample_profile", "razorpay"
        )

        self.post_api_mock.assert_called_with(
            "api/v1/payments/qr/code/closure",
            data=self.qr_code_closure_dict,
            send_request_id=True,
            add_bearer_token=True,
        )

        self.get_api_mock.assert_called_with(
            "api/v1/payments/result",
            headers={"X-Mercury-Request-Id": "random_string"},
            send_request_id=False,
            add_bearer_token=True,
        )

        self.assertEqual(response[1]["status"], "SUCCESS")

    def test_qr_code_closure_exception_raised_if_status_code_error(self):
        mock_response = mock.MagicMock()
        mock_response.status_code = 401
        self.post_api_mock.return_value = ("random_string", mock_response)
        with self.assertRaises(Exception):
            self.client.qr_code_closure(
                self.qr_code_closure_obj, "test", "some_provider"
            )

    def test_entire_payment_plan_flow(self):
        mock_post_response = mock.MagicMock()
        mock_post_response.status_code = 201
        mock_get_response = mock.MagicMock()
        mock_get_response.status_code = 200
        mock_get_response.json = mock.MagicMock()
        mock_get_response.json.return_value = {"status": "SUCCESS"}
        self.post_api_mock.return_value = ("random_string", mock_post_response)
        self.get_api_mock.return_value = ("random_string", mock_get_response)

        response = self.client.generate_payment_plan(
            self.plan_creation_obj, "sample_profile", "razorpay"
        )

        self.post_api_mock.assert_called_with(
            "api/v1/payments/plan/create",
            data=self.plan_creation_dict,
            send_request_id=True,
            add_bearer_token=True,
        )

        self.get_api_mock.assert_called_with(
            "api/v1/payments/result",
            headers={"X-Mercury-Request-Id": "random_string"},
            send_request_id=False,
            add_bearer_token=True,
        )

        self.assertEqual(response[1]["status"], "SUCCESS")

    def test_plan_creation_request_exception_raised_if_status_code_error(self):
        mock_response = mock.MagicMock()
        mock_response.status_code = 401
        self.post_api_mock.return_value = ("random_string", mock_response)
        with self.assertRaises(Exception):
            self.client.generate_payment_plan(
                self.plan_creation_obj, "test", "some_provider"
            )

    def test_entire_subscription_creation_flow(self):
        mock_post_response = mock.MagicMock()
        mock_post_response.status_code = 201
        mock_get_response = mock.MagicMock()
        mock_get_response.status_code = 200
        mock_get_response.json = mock.MagicMock()
        mock_get_response.json.return_value = {"status": "SUCCESS"}
        self.post_api_mock.return_value = ("random_string", mock_post_response)
        self.get_api_mock.return_value = ("random_string", mock_get_response)

        response = self.client.create_payment_subscription(
            self.subscription_creation_obj, "sample_profile", "razorpay"
        )

        self.post_api_mock.assert_called_with(
            "api/v1/payments/subscription/create",
            data=self.subscription_creation_dict,
            send_request_id=True,
            add_bearer_token=True,
        )

        self.get_api_mock.assert_called_with(
            "api/v1/payments/result",
            headers={"X-Mercury-Request-Id": "random_string"},
            send_request_id=False,
            add_bearer_token=True,
        )

        self.assertEqual(response[1]["status"], "SUCCESS")

    def test_subscription_request_exception_raised_if_status_code_error(self):
        mock_response = mock.MagicMock()
        mock_response.status_code = 401
        self.post_api_mock.return_value = ("random_string", mock_response)
        with self.assertRaises(Exception):
            self.client.create_payment_subscription(
                self.subscription_creation_obj, "test", "some_provider"
            )

    def test_entire_subscription_fetch_flow(self):
        mock_post_response = mock.MagicMock()
        mock_post_response.status_code = 201
        mock_get_response = mock.MagicMock()
        mock_get_response.status_code = 200
        mock_get_response.json = mock.MagicMock()
        mock_get_response.json.return_value = {"status": "SUCCESS"}
        self.post_api_mock.return_value = ("random_string", mock_post_response)
        self.get_api_mock.return_value = ("random_string", mock_get_response)

        response = self.client.fetch_subscription(
            self.subscription_fetch_obj, "sample_profile", "razorpay"
        )

        self.post_api_mock.assert_called_with(
            "api/v1/payments/subscription/fetch",
            data=self.subscription_fetch_dict,
            send_request_id=True,
            add_bearer_token=True,
        )

        self.get_api_mock.assert_called_with(
            "api/v1/payments/result",
            headers={"X-Mercury-Request-Id": "random_string"},
            send_request_id=False,
            add_bearer_token=True,
        )

        self.assertEqual(response[1]["status"], "SUCCESS")

    def test_fetch_subs_request_exception_raised_if_status_code_error(self):
        mock_response = mock.MagicMock()
        mock_response.status_code = 401
        self.post_api_mock.return_value = ("random_string", mock_response)
        with self.assertRaises(Exception):
            self.client.fetch_subscription(
                self.subscription_fetch_obj, "test", "some_provider"
            )

    def test_entire_payment_link_flow(self):
        mock_post_response = mock.MagicMock()
        mock_post_response.status_code = 201
        mock_get_response = mock.MagicMock()
        mock_get_response.status_code = 200
        mock_get_response.json = mock.MagicMock()
        mock_get_response.json.return_value = {"status": "SUCCESS"}
        self.post_api_mock.return_value = ("random_string", mock_post_response)
        self.get_api_mock.return_value = ("random_string", mock_get_response)

        response = self.client.generate_payment_link(
            self.payment_link_obj, "sample_profile", "razorpay"
        )

        self.post_api_mock.assert_called_with(
            "api/v1/payments/payment/link/create",
            data=self.payment_link_dict,
            send_request_id=True,
            add_bearer_token=True,
        )

        self.get_api_mock.assert_called_with(
            "api/v1/payments/result",
            headers={"X-Mercury-Request-Id": "random_string"},
            send_request_id=False,
            add_bearer_token=True,
        )

        self.assertEqual(response[1]["status"], "SUCCESS")

    def test_payment_link_request_exception_raised_if_status_code_error(self):
        mock_response = mock.MagicMock()
        mock_response.status_code = 401
        self.post_api_mock.return_value = ("random_string", mock_response)
        with self.assertRaises(Exception):
            self.client.fetch_subscription(
                self.payment_link_obj, "test", "some_provider"
            )

    def test_cancel_payment_link_flow(self):
        mock_post_response = mock.MagicMock()
        mock_post_response.status_code = 201
        mock_get_response = mock.MagicMock()
        mock_get_response.status_code = 200
        mock_get_response.json = mock.MagicMock()
        mock_get_response.json.return_value = {"status": "SUCCESS"}
        self.post_api_mock.return_value = ("random_string", mock_post_response)
        self.get_api_mock.return_value = ("random_string", mock_get_response)

        response = self.client.cancel_payment_link(
            self.payment_link_closure_obj, "sample_profile", "razorpay"
        )

        self.post_api_mock.assert_called_with(
            "api/v1/payments/payment/link/cancel",
            data=self.payment_link_closure_dict,
            send_request_id=True,
            add_bearer_token=True,
        )

        self.get_api_mock.assert_called_with(
            "api/v1/payments/result",
            headers={"X-Mercury-Request-Id": "random_string"},
            send_request_id=False,
            add_bearer_token=True,
        )

        self.assertEqual(response[1]["status"], "SUCCESS")

    def test_cancel_payment_link_exception_raised_if_status_code_error(self):
        mock_response = mock.MagicMock()
        mock_response.status_code = 401
        self.post_api_mock.return_value = ("random_string", mock_response)
        with self.assertRaises(Exception):
            self.client.cancel_payment_link(
                self.payment_link_closure_obj, "test", "some_provider"
            )

    def test_payments_result_response(self):
        mock_response = mock.MagicMock()
        mock_response.status_code = 200
        self.get_api_mock.return_value = ("random_string", mock_response)
        self.client.get_payment_response("random_string")

        self.get_api_mock.assert_called_with(
            "api/v1/payments/result",
            headers={"X-Mercury-Request-Id": "random_string"},
            send_request_id=False,
            add_bearer_token=True,
        )

    def test_payment_result_exception_raised_if_status_code_error(self):
        mock_response = mock.MagicMock()
        mock_response.status_code = 401
        self.get_api_mock.return_value = ("random_string", mock_response)
        with self.assertRaises(Exception):
            self.client.get_payment_result("random_string")

    def test_payment_result_response_api_succeeds_if_status_code_success(self):
        mock_response = mock.MagicMock()
        mock_response.status_code = 200
        mock_response.json = mock.MagicMock()
        mock_response.json.return_value = {"status": "SUCCESS"}
        self.get_api_mock.return_value = ("random_string", mock_response)

        request_id, result = self.client.get_payment_result("random_string")
        self.assertEqual(request_id, "random_string")
        self.assertEqual(result["status"], "SUCCESS")
